@@ -71,7 +71,7 @@ const tests = [
       'ab = "aaaa"', ENTER,
       '555 - 909', ENTER,
       '{key : {key2 :[] }}', ENTER,
-      'Array(100).fill(1)', ENTER
+      'Array(100).fill(1)', ENTER,
     ],
     expected: [],
     clean: false
@@ -99,7 +99,7 @@ const tests = [
       SEARCH_BACKWARDS, // 15
       SEARCH_FORWARDS,
       ESCAPE,           // 17
-      ENTER
+      ENTER,
     ],
     // A = Cursor n up
     // B = Cursor n down
@@ -179,7 +179,7 @@ const tests = [
       '\x1B[1G', '\x1B[0J',
       prompt,
       '\x1B[3G',
-      '\r\n'
+      '\r\n',
     ],
     clean: false
   },
@@ -207,14 +207,12 @@ const tests = [
       UP,                // 15
       DOWN,
       SEARCH_FORWARDS,   // 17
-      '\n'
+      '\n',
     ],
     expected: [
       '\x1B[1G', '\x1B[0J',
       prompt, '\x1B[3G',
-      'f', 'u', ' // nction',
-      '\x1B[5G', '\x1B[0K',
-      '\nbck-i-search: _', '\x1B[1A', '\x1B[5G',
+      'f', 'u', '\nbck-i-search: _', '\x1B[1A', '\x1B[5G',
       '\x1B[3G', '\x1B[0J',
       '{key : {key2 :[] }}\nbck-i-search: }_', '\x1B[1A', '\x1B[21G',
       '\x1B[3G', '\x1B[0J',
@@ -262,10 +260,10 @@ const tests = [
       '-1\n',
       '\x1B[1G', '\x1B[0J',
       prompt, '\x1B[3G',
-      '\r\n'
+      '\r\n',
     ],
     clean: false
-  }
+  },
 ];
 const numtests = tests.length;
 
